(**************************************************************************)
(*  ProuveParser: parser frontend for the PROUV language                 *)
(*                                                                        *)
(*  Copyright (C) 2004, 2005 Ralf Treinen                                 *)
(*                                                                        *)
(*  This library is free software; you can redistribute it and/or         *)
(*  modify it under the terms of the GNU Lesser General Public            *)
(*  License as published by the Free Software Foundation; either          *)
(*  version 2.1 of the License, or (at your option) any later version.    *)
(*                                                                        *)
(*  This library is distributed in the hope that it will be useful,       *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU     *)
(*  Lesser General Public License for more details.                       *)
(*                                                                        *)
(*  You should have received a copy of the GNU Lesser General Public      *)
(*  License along with this library; if not, write to the Free Software   *)
(*  Foundation, Inc., 59 Temple Place, Suite 330,                         *)
(*  Boston, MA  02111-1307  USA                                           *)
(*                                                                        *)
(*  Ralf Treinen <treinen@lsv.ens-cachan.fr>                              *)
(*  Laboratoire Spcification et Vrification                             *)
(*  cole Normale Suprieure de Cachan, CNRS UMR 8643                     *)
(*  61, avenue du Prsident Wilson                                        *)
(*  94235 CACHAN Cedex - France                                           *)
(**************************************************************************)

open Types

(* universal type of expressions *)
type expression =
    Ident of string
  | Int of int
  | Apply of string * expression list
  | Infix of string * expression * expression
  | Inverse of expression
  | Tuple of expression list
  | Liste of expression list
  | AssocList of (expression*expression) list

type pattern = expression list


(************************** Conversion to strings ***********************)

let rec string_of_expression = function
    Ident s -> s
  | Int i -> (string_of_int i)
  | Apply(f,l) -> f ^  
      (match l with
	   [] -> ""
	 | _  -> "(" ^ (string_of_expression_list l) ^ ")")
  | Infix(op,e1,e2) ->
      "(" ^ (string_of_expression e1) ^ " " ^ op ^
      " " ^ (string_of_expression e2) ^ ")"
  | Inverse e -> "inv(" ^ (string_of_expression e) ^ ")"
  | Tuple l -> "[" ^ (string_of_expression_list l) ^ "]"
  | Liste l -> "[[" ^ (string_of_expression_list l) ^ "]]"
  | AssocList l -> "{" ^ (string_of_expression_assoclist l) ^ "}"
and string_of_expression_list = function
    [] -> ""
  | h::r ->
      List.fold_left
	(fun s e -> s^", "^(string_of_expression e) )
	(string_of_expression h)
	r
and string_of_expression_assoclist = function
    [] -> ""
  | (h1,h2)::r ->
      List.fold_left
	(fun s (e1,e2) ->
	   s^", "^(string_of_expression e1)^": "^(string_of_expression e2) )
	((string_of_expression h1)^": "^(string_of_expression h2))
	r

let string_of_pattern pat =
  Utils.string_of_string_list "; "
    (List.map string_of_expression pat) 

(************ renaming of variables *************************************)

let rec rename_expr alpha = function
	 Ident x -> 
		(try
		  Ident (List.assoc x alpha)
		 with
			  Not_found -> Ident x)
  | Int i -> Int i
  | Infix (op,e1,e2) -> Infix(op, rename_expr alpha e1, rename_expr alpha e2)
  | Inverse e -> Inverse (rename_expr alpha e)
  | Tuple l -> Tuple (List.map (rename_expr alpha) l) 
  | Liste l -> Liste (List.map (rename_expr alpha) l)
  | Apply (f,l) -> Apply(f,List.map (rename_expr alpha) l) 
  | AssocList l ->
      AssocList (List.map
		   (function (l,r) ->
		      (rename_expr alpha l, rename_expr alpha r))
						 l)

let rename_pattern ren = List.map (rename_expr ren)

(************   restricted forms of expressions   ***********************)

let is_l_expression = function
    Ident _ | Apply(_,_) -> true
  | _ -> false

(****************  collect integer constants  *****************************)

let rec ic = function
    Ident _ -> []
  | Int i -> [i]
  | Inverse e -> ic e
  | AssocList l ->
      (List.flatten (List.map ic (List.map fst l))) @
      (List.flatten (List.map ic (List.map snd l)))
  | Apply (_,el) | Tuple el | Liste el -> List.flatten (List.map ic el)
  | Infix(_,e1,e2) -> (ic e1) @ (ic e2)

let int_constants e = Utils.singlyfy (ic e)


(*************************** Typing *************************************)
exception Exc_typeexpression of string
let rec type_expr sign env skenv r_env = function
    Int _ -> Typeconst "int"
  | Tuple l -> Typetuple (List.map (type_expr sign env skenv r_env) l)
  | Ident s ->
      (try
	snd (List.assoc s env)
	  (* environemnts contain only arrow-free types *)
      with
	  Not_found -> 
	    try
	      let t = List.assoc s sign
	      in if arrow_free t
		then t
		else raise
		  (Exc_typeexpression
		     ("the identifier \"" ^ s ^ "\" is used without arguments "
		      ^ "but is declared as of functional type "
		      ^ (string_of_type_expression t)))
	    with
		Not_found -> raise
		  (Exc_typeexpression
		     ("the identifier \"" ^ s ^ "\" is not declared, " ^
		      "neither as a variable nor as a constant operation")))
  | Apply(f,l) ->
      (try
	 let  (from_type_list,_,result_type) = List.assoc f skenv 
	 in let arguments_type_list =
	     List.map (type_expr sign env skenv r_env) l
	 in
	   (try
	      apply_type_subst
		(type_match 
		   (Typetuple from_type_list)
		   (Typetuple arguments_type_list))
		result_type
	    with
		Nomatch_type -> raise
		  (Exc_typeexpression
		     ("the identifier \"" ^ f ^ "\" is declared as a " ^
		      "function with argument type " ^
		      (string_of_type_expression_list from_type_list) ^
		      ", which does not match the type " ^
		      (string_of_type_expression_list arguments_type_list))))
       with Not_found ->
	 (try
	    match List.assoc f sign with
		Typearrow(from_type, result_type) ->
		  let arguments_type =
		    if List.length l = 1
		    then type_expr sign env skenv r_env (List.hd l)
		    else (Typetuple
			    (List.map (type_expr sign env skenv r_env) l))
		  in
		    (try
		       apply_type_subst
			 (type_match from_type arguments_type) result_type
		     with
			 Nomatch_type -> raise
			   (Exc_typeexpression
			      ("the identifier \"" ^ f ^
			       "\" is declared as a " ^
			       "function with argument type " ^
			       (string_of_type_expression from_type) ^
			       ", which does not match the type " ^
			       (string_of_type_expression arguments_type))))
	      |  ftype -> raise
		   (Exc_typeexpression
		      ("the identifier \"" ^ f ^ "\"is used like a function " ^
		       "but has non-functional type " ^
		       (string_of_type_expression ftype)))
	  with
	      Not_found -> raise
		(Exc_typeexpression
		   ("the identifier \"" ^ f ^ "\" is used like a function "^
		    "but is not declared"))))
  | Infix(op,e1,e2) ->
      let left_type = type_expr sign env skenv r_env e1
      in
      if op = "." &&
	(match left_type
	 with Typeconst(n) -> List.mem_assoc n r_env
	   | _ -> false)
      then (* we have an access to a toplevel identifier of a process *)
	match e2 with
	    Ident vid -> 
	      (match left_type with
		   Typeconst rolename ->
		     (try snd (List.assoc vid (List.assoc rolename r_env))
		      with
			  Not_found ->
			    raise
			      (Exc_typeexpression
				 ("the identifier \""^vid^"\" is not a " ^
				    "toplevel identifier of role \""^rolename^
				    "\"")))
		 | _ -> failwith "this cannot happen")
	  | _ -> raise
	      (Exc_typeexpression
		 ((string_of_expression e2) ^ "is not an identifier"))
      else (* not an access to a toplevel identifier of a process *)
      let right_type = type_expr sign env skenv r_env e2
      in
	(try
	   match List.assoc op sign with
	       Typearrow(from,res) ->
		 (try 
		    apply_type_subst
		      (type_match from
			 (Typetuple [left_type; right_type]))
		      res
		  with
		      Nomatch_type -> raise
		  (Exc_typeexpression
		     ("the identifier \"" ^ op ^ "\" is declared as an " ^
		      "infix operator with argument argument type " ^
		      (string_of_type_expression from) ^
		      " but is applied to arguments of respective types " ^
		      (string_of_type_expression left_type) ^
		      " and " ^
		      (string_of_type_expression right_type))))

	     |  ftype -> raise (Exc_typeexpression
				  (op ^ "is used as an infix operator but has " ^
				     "non-functional type" ^
				     (string_of_type_expression ftype)))
	 with
	     Not_found ->
	       raise (Exc_typeexpression
			("infix operator " ^ op ^ " is not declared")))
  | Inverse e ->
      (match type_expr sign env skenv r_env e with
	  Typeconst "pubkey" -> Typeconst "privkey"
	| Typeconst "privkey" -> Typeconst "pubkey"
	| Typeconst "message" -> Typeconst "message"
	| othertype -> raise (Exc_typeexpression
				("Argument of Inverse operation of type "
				 ^ (string_of_type_expression othertype))))
  | Liste l as complete_l -> 
	 Typeapply("list",
		   [List.fold_left sup_type 
		      Typeundef
		      (List.map (type_expr sign env skenv r_env ) l)])
  | AssocList l as complete_l ->
	 Typeapply("table",
		   [List.fold_left sup_type
		      Typeundef
		      (List.map (function (x,_) ->
				   type_expr sign env skenv r_env x) l);
		    List.fold_left sup_type
		      Typeundef
		      (List.map (function (_,x) ->
				   type_expr sign env skenv r_env x) l)
		   ])

	    
	
