(**************************************************************************)
(*  ProuveParser: parser frontend for the PROUV language                 *)
(*                                                                        *)
(*  Copyright (C) 2004, 2005 Ralf Treinen                                 *)
(*                                                                        *)
(*  This library is free software; you can redistribute it and/or         *)
(*  modify it under the terms of the GNU Lesser General Public            *)
(*  License as published by the Free Software Foundation; either          *)
(*  version 2.1 of the License, or (at your option) any later version.    *)
(*                                                                        *)
(*  This library is distributed in the hope that it will be useful,       *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU     *)
(*  Lesser General Public License for more details.                       *)
(*                                                                        *)
(*  You should have received a copy of the GNU Lesser General Public      *)
(*  License along with this library; if not, write to the Free Software   *)
(*  Foundation, Inc., 59 Temple Place, Suite 330,                         *)
(*  Boston, MA  02111-1307  USA                                           *)
(*                                                                        *)
(*  Ralf Treinen <treinen@lsv.ens-cachan.fr>                              *)
(*  Laboratoire Spcification et Vrification                             *)
(*  cole Normale Suprieure de Cachan, CNRS UMR 8643                     *)
(*  61, avenue du Prsident Wilson                                        *)
(*  94235 CACHAN Cedex - France                                           *)
(**************************************************************************)

(* formulas *)

open Expressions
open Declarations
open Types

type formula =
    Ftrue
  | Fatom of expression
  | Fsecret of expression
  | Fprocessat of string * string
  | Fscenarioat of string
  | Fand of formula * formula
  | For of formula * formula
  | Fimplies of formula * formula
  | Fnot of formula
  | Fforall of declaration list * formula
  | Fexists of declaration list * formula
  | Funique of declaration list * formula
  | Fbox of formula

let rec is_quantifier_free = function
    Ftrue | Fatom _ | Fsecret _ | Fprocessat _ | Fscenarioat _ -> true
  | Fnot f -> is_quantifier_free f
  | Fand(f1,f2) | For(f1,f2) | Fimplies(f1,f2) ->
      is_quantifier_free f1 && is_quantifier_free f2
  | _ -> false

(* universal closure of a quantifier-free formula *)
let rec is_pi1 = function
    Fforall (_,f) -> is_pi1 f
  | f -> is_quantifier_free f

(* existential closure of a quantifier-free formula *)
let rec is_sigma1 = function
    Fexists (_,f) | Funique(_,f) -> is_sigma1 f
  | f -> is_quantifier_free f

(* positive boolean combination of sigma1-formulas *)
let rec is_posbool_sigma1 = function
    Fand(f1,f2) | For(f1,f2) -> is_posbool_sigma1 f1 && is_posbool_sigma1 f2
  | f -> is_sigma1 f

(* safety formula : universal closure of an implication b -> g where *)
(* - b is quantifier-free                                            *)
(* - g is a  positive boolean combination of sigma1-formulas         *)
let rec is_safety_formula = function
    Fforall (_,f) -> is_safety_formula f
  | Fimplies (b,g) -> is_quantifier_free b && is_posbool_sigma1 g
  | _ -> false

(*********** Type checking *******************************************)

exception Exc_formula of string

let rec check_types sign env r_env r_cp s_cp = function
    Ftrue -> ()
  | Fatom e ->
      let et = Expressions.type_expr sign env [] r_env e 
      in if et <> Typeconst "bool"
	then raise (Exc_formula
		      ("atomic formula of type " ^ 
			 (string_of_type_expression et) ^
			 "instead of type bool."))
  | Fsecret e -> ignore (Expressions.type_expr sign env [] r_env e)
  | Fnot f | Fbox f -> check_types sign env r_env r_cp s_cp f
  | Fand(f1,f2) | For(f1,f2) | Fimplies(f1,f2) ->
      check_types sign env r_env r_cp s_cp f1;
      check_types sign env r_env r_cp s_cp f2
  | Funique(dl,f) | Fexists(dl,f) | Fforall(dl,f) ->
      check_types sign (dl@env) r_env r_cp s_cp f
  | Fprocessat(pid,point) ->
      begin 
	try 
	  let (kind,pid_type) = List.assoc pid env
	  in if kind <> ProcessKind
	    then raise (Exc_formula
			  ("variable \""^pid^"\" in AT statement " ^
			     "of type \"" ^
			     (string_of_type_expression pid_type)^
			     "\" instead of a role type"))
	    else begin
	      match pid_type with
		  Typeconst(role) ->
		    if not (List.mem point (List.assoc role r_cp))
		    then raise
		      (Exc_formula
			 ("variable \""^pid^"\" in AT statement " ^
			    "at point \""^point^"\" but this is not a " ^
			    "control point of role \""^role^"\""))
		| _ -> failwith
		    "this cannot happen: processkind and no typeconst"
	    end
	with
	    Not_found -> raise (Exc_formula
				  ("variable \""^pid^"\" in AT statement " ^
				     "is not bound"))
      end
  | Fscenarioat(point) -> 
      if not (List.mem point s_cp)
      then raise (Exc_formula
		    ("AT statement for scenario at point \""^point^
		       "\" but this is not a valid checkpoint of the "^
		       "scenario"))
	    
let check_types_formula sign env r_env r_checkpoints s_checkpoints f = 
  try check_types sign env r_env r_checkpoints s_checkpoints f
  with 
      Expressions.Exc_typeexpression s -> raise (Exc_formula s)
