(**************************************************************************)
(*  ProuveParser: parser frontend for the PROUV language                 *)
(*                                                                        *)
(*  Copyright (C) 2004, 2005 Ralf Treinen                                 *)
(*                                                                        *)
(*  This library is free software; you can redistribute it and/or         *)
(*  modify it under the terms of the GNU Lesser General Public            *)
(*  License as published by the Free Software Foundation; either          *)
(*  version 2.1 of the License, or (at your option) any later version.    *)
(*                                                                        *)
(*  This library is distributed in the hope that it will be useful,       *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU     *)
(*  Lesser General Public License for more details.                       *)
(*                                                                        *)
(*  You should have received a copy of the GNU Lesser General Public      *)
(*  License along with this library; if not, write to the Free Software   *)
(*  Foundation, Inc., 59 Temple Place, Suite 330,                         *)
(*  Boston, MA  02111-1307  USA                                           *)
(*                                                                        *)
(*  Ralf Treinen <treinen@lsv.ens-cachan.fr>                              *)
(*  Laboratoire Spcification et Vrification                             *)
(*  cole Normale Suprieure de Cachan, CNRS UMR 8643                     *)
(*  61, avenue du Prsident Wilson                                        *)
(*  94235 CACHAN Cedex - France                                           *)
(**************************************************************************)

open Types

type simple_signature = (string * type_expression ) list

type signature = simple_signature * simple_signature

let flatten_signature l t = List.map (function x -> (x,t)) l

let is_empty = function [] -> true | _ -> false

(**************** printing ***********************************************)
(* print a simple signature without keywords *)
let print_simple oc = function
    [] -> ()
  | signature ->
      List.iter
		  (function (fname, ftype) ->
			  output_string oc "  ";
			  output_string oc  fname;
			  output_string oc ": ";
			  output_string oc (string_of_type_expression ftype);
			  output_string oc  ";\n")
		  signature

(* print a signature with all necessary keywords *)
let print_signature oc (psig, hsig) =
  let os = output_string oc
  in 
    if not(is_empty psig && is_empty hsig )
    then begin
      os "signature\n";
      print_simple oc psig;
      if not (is_empty hsig)
      then begin
	os "hashfunctions\n";
	print_simple oc hsig
      end;
      os "end\n"
    end

(************************************************************************)
      
let list_of_constants (psig,hsig) =
  List.map fst
    (List.filter 
       (function (s,t) -> match t with
	    Typeconst _ -> true
	  | _ -> false)
       (psig@hsig))


exception Exc_signatures of string 

(* check that no symbol is declared both in psig and in hsig *)
let check_disjoint psig hsig =
  let pl = List.map fst psig
  and hl = List.map fst hsig
  in 
	 try
		let double = List.find (function x -> List.mem x pl) hl
		in raise (Exc_signatures
						("\"" ^ double ^ "\" declared both as " ^
							"decomposable function symbol and as hash function"))
	 with
		  Not_found -> ()

(* try to construct the signature obtained by adjoining to sign the *)
(* declaration "f:te". Error when te is not a well_formed type      *)
(* expression or when f is already defined in sign.                 *)
let add_to_sig sign (f, te) =
  if List.mem_assoc f sign
  then raise (Exc_signatures ("\"" ^ f ^ "\" is already declared."))
  else try
    well_formed te;
    (f,te)::sign 
  with
      Wellformed_type s -> raise (Exc_signatures s)
   
let construct_signature (psig,hsig) =
  check_disjoint psig hsig;
  (List.fold_left add_to_sig Defaultsignature.default_signature psig,
	hsig)



