(**************************************************************************)
(*  ProuveParser: parser frontend for the PROUV language                 *)
(*                                                                        *)
(*  Copyright (C) 2004, 2005 Ralf Treinen                                 *)
(*                                                                        *)
(*  This library is free software; you can redistribute it and/or         *)
(*  modify it under the terms of the GNU Lesser General Public            *)
(*  License as published by the Free Software Foundation; either          *)
(*  version 2.1 of the License, or (at your option) any later version.    *)
(*                                                                        *)
(*  This library is distributed in the hope that it will be useful,       *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU     *)
(*  Lesser General Public License for more details.                       *)
(*                                                                        *)
(*  You should have received a copy of the GNU Lesser General Public      *)
(*  License along with this library; if not, write to the Free Software   *)
(*  Foundation, Inc., 59 Temple Place, Suite 330,                         *)
(*  Boston, MA  02111-1307  USA                                           *)
(*                                                                        *)
(*  Ralf Treinen <treinen@lsv.ens-cachan.fr>                              *)
(*  Laboratoire Spcification et Vrification                             *)
(*  cole Normale Suprieure de Cachan, CNRS UMR 8643                     *)
(*  61, avenue du Prsident Wilson                                        *)
(*  94235 CACHAN Cedex - France                                           *)
(**************************************************************************)

open Specs

exception Exc_specparser of string

(*****************************************************************)
(*    Constructing the abstract syntax                           *)
(*****************************************************************)
      
let parse_spec infilename =
  try
    let inbuf = Lexing.from_channel (open_in infilename)
    in
      (try
	 begin
	   Lexer.register_filename inbuf infilename;
	   let
	       ((psig,hsig), axioms, rolelist, vars, scenario) as spec =
	     Parser.spec Lexer.next_token inbuf
	   in
	   let 
	       (cpsig,chsig) = Signatures.construct_signature (psig,hsig)
	   and
	       role_declarations = Roles.gather_parameter_declarations rolelist
	   in let
		 complete_sig = cpsig@chsig
	   in
	     Axioms.check_axioms complete_sig axioms;
	     Roles.check_types_roles complete_sig rolelist;
	     Scenarios.check_types_scenario
	       complete_sig role_declarations vars scenario;
	     spec
	 end
       with
	   Lexer.Lex_error(s,p) ->
	     raise (Exc_specparser
		      ("Lexer error on character ``" ^ s ^ "'' at " ^ 
			 (Lexer.string_of_lexposition p )))
	 | Parsing.Parse_error ->
	     raise (Exc_specparser
		      ("Parsing error before " ^
			 (Lexer.string_of_lexposition
			    (Lexing.lexeme_start_p inbuf))))
	 | Signatures.Exc_signatures s ->
	     raise (Exc_specparser
		      ("File " ^ infilename ^
			 ": Cannot construct the signature because " ^ s))
	 | Declarations.Exc_declarations s ->
	     raise (Exc_specparser
		      ("File " ^ infilename ^
			 ": Declaration incorrect because " ^ s))
	 | Axioms.Exc_axioms s ->
	     raise (Exc_specparser
		      ("File " ^ infilename ^
			 ": Axioms incorrect because " ^ s)	)
	 | Roles.Exc_roles s ->
	     raise (Exc_specparser
		      ("File " ^ infilename ^
			 ": Role list incorrect because " ^ s))
	 | Scenarios.Exc_scenario s ->
	     raise (Exc_specparser
		      ("File " ^ infilename ^
			 ": Scenario incorrect because " ^ s))
	 | Intervals.Exc_intervals s ->
	     raise (Exc_specparser
		      ("File " ^ infilename ^
			 ": Interval incorrect because " ^ s)))
  with
      Sys_error s -> raise (Exc_specparser ("cannot open file " ^ infilename ^
					    " for reading because " ^ s))
